/* =====================================================================
 *
 * Copyright 2000 The Board of Trustees of the University of Illinois
 * All Rights Reserved
 *
 * Author(s): Volodymyr Kindratenko @ NCSA
 * Last Modified: January 2000
 *
 * calib.c - VRCO trackd calibration filter
 *
 * =====================================================================*/

#include "calib.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* uses TrCalibrLib ver. 2.0 */
#include "calibfile.h"
#include "quaternions.h"
#include "polyfit.h"
#include "interpolation.h"

/* calibration data structure */
struct _init 
{
  /* only one of them is not NULL and it defines the type of calibration */
  struct fpolynomials *fpol;
  struct rotcor *rc;

  int param;  /* needed for the interpolation */
};


/* The params is an array of pointers to character. If you write
  "DeviceOption devName calibration your.so your_data 1 radians 25",
  *params points to "your_data", *(params+1) points to "1",
  *(params+2) points to "radians" and *(params+3) points to "25".
  numParams is the number of parameters, it is 4 in this case.
  The return value is the pointer to the data structure casted to
  void *. */
extern "C" LINKAGE void* calib_init(const char* const* params, int numParams)
{
  /* calibration algorithms */
#define FIT 1011
#define ITP 1012

  int algtm, param, rottype;

  /* parameters */
  if (numParams != 4) {
    printf("WARNING: calibration filter is not attached (wrong number of arguments: %i\n", numParams);
    return NULL;
  }

  /* calibration algorithm */
  if (strcmp(*params, "FIT") == 0) algtm = FIT;
  else if (strcmp(*params, "ITP") == 0) algtm = ITP;
  else {
    printf("WARNING: calibration filter is not attached (unknown calibration procedure specified: %s\n", *params);
    return NULL;
  }
  
  /* fit order or weight type */
  param = atoi(*(params+1));
  if ( (algtm == FIT) && ( (param < 2) || (param > 7) ) ) {
    printf("WARNING: calibration filter is not attached (unsupported polynomial order: %s\n", *(params+1));
    return NULL;
  }
  else if ( (algtm == ITP) && ( (param < 1) || (param > 3) ) ) {
    printf("WARNING: calibration filter is not attached (unsupported weights type: %s\n", *(params+1));
    return NULL;
  }

  /* angles in the calibration file */
  if (strcmp(*(params+3), "RADIANS") == 0) rottype = RADIANS;
  else if (strcmp(*(params+3), "DEGREES") == 0) rottype = DEGREES;
  else {
    printf("WARNING: calibration filter is not attached (unsupported type of angular units: %s\n", *(params+3));
    return NULL;
  }

  /* calibration table */
  struct calibdata ctable;

  /* read the calibration file */
  if (!readcalibfile((char *)*(params+2), &ctable, rottype))
    {
      printf("WARNING: calibration filter is not attached (unable to read calibration file: %s\n", *(params+2));
      return NULL;
    }

  /* data structure to be returned by the function */
  struct _init* data = new struct _init;
  data->fpol = NULL;
  data->rc = NULL;
  data->param = param;

  /* make and initialize an appropriate sub-structure */
  int res = 0;
  if (algtm == FIT) {
    data->fpol = new struct fpolynomials;
    res = fit_initialize(&ctable, param, data->fpol);
  } else {
    data->rc = new struct rotcor;
    res = itp_initialize(&ctable, data->rc);
  }

  /* remove calibration file data structutre */
  deletecalibdata(&ctable);

  /* if unable to initialize, remove the data and return NULL */
  if (!res) {
    if (data->fpol) delete data->fpol;
    if (data->rc) delete data->rc;
    delete data;
    printf("WARNING: calibration filter is not attached (unable to initialize)\n");
    return NULL;
  }
  
  printf("Calibration filter is attached: %s %i %s %s\n",
	 *params, data->param, *(params+2), *(params+3));

  return (void *)data;
}

/* data is the data structure that you got from your_data.
  trackerData is the array of 6 float valuables, x, y, z, azim, elev,
  and roll. */
extern "C" LINKAGE void calib_evaluate(void *data, float *trackerData)
{
  float x_in = trackerData[0];  /* X */
  float y_in = trackerData[1];  /* Y */
  float z_in = trackerData[2];  /* Z */
  float x_out, y_out, z_out;
  float az_in = trackerData[3]/180.0f*3.1416f; /* Azim */
  float el_in = trackerData[4]/180.0f*3.1416f; /* Elev */
  float rl_in = trackerData[5]/180.0f*3.1416f; /* Roll */
  float az_out, el_out, rl_out;

  if (((struct _init*)data)->fpol) { // polynomial fit
    fit_evaluate(((struct _init*)data)->fpol,
		 x_in, y_in, z_in, az_in, el_in, rl_in,
		 &x_out, &y_out, &z_out, &az_out, &el_out, &rl_out);
    az_out = az_out/3.1416f*180.0f;
    el_out = el_out/3.1416f*180.0f;
    rl_out = rl_out/3.1416f*180.0f;
#if DEBUG
    printf("DEBUG: calib_evaluate: polynomial fit\n");
#endif
  }
  else if (((struct _init*)data)->rc) { // interpolation
    itp_evaluate(((struct _init*)data)->rc, ((struct _init*)data)->param,
		 x_in, y_in, z_in, az_in, el_in, rl_in,
		 &x_out, &y_out, &z_out, &az_out, &el_out, &rl_out);
    az_out = az_out/3.1416f*180.0f;
    el_out = el_out/3.1416f*180.0f;
    rl_out = rl_out/3.1416f*180.0f;
#if DEBUG
    printf("DEBUG: calib_evaluate: interpolation\n");
#endif
  }
  else { // should never be here!!!
    x_out = x_in; y_out = y_in; z_out = z_in;
    az_out = az_in; el_out = el_in; rl_out = rl_in;
#if DEBUG
    printf("DEBUG: calib_evaluate: Something went terribly wrong!\n");
#endif
  }

  trackerData[0] = x_out;  /* X */
  trackerData[1] = y_out;  /* Y */
  trackerData[2] = z_out;  /* Z */
  trackerData[3] = az_out; /* Azim */
  trackerData[4] = el_out; /* Elev */
  trackerData[5] = rl_out; /* Roll */
  
#if DEBUG
  printf("DEBUG: calib_evaluate:  in: %2.2f %2.2f %2.2f %2.2f %2.2f %2.2f\n",
	 x_in, y_in, z_in, az_in, el_in, rl_in);
  printf("DEBUG: calib_evaluate: out: %2.2f %2.2f %2.2f %2.2f %2.2f %2.2f\n",
	 x_out, y_out, z_out, az_out, el_out, rl_out);
#endif
}

/* You have to delete the memory you assgined by init. */
extern "C" LINKAGE void calib_exit(void *data)
{
  if (((struct _init*)data)->fpol) delete ((struct _init*)data)->fpol;
  if (((struct _init*)data)->rc) delete ((struct _init*)data)->rc;
  delete (struct _init*)data;

  printf("Calibration filter is destroyed\n");
}

